from pawnlib.config import pawnlib_config as pawn
from pawnlib.builder.generator import AppGenerator, generate_banner
from pawnlib.__version__ import __version__ as _version
import argparse

from rich.console import Console
from rich.syntax import Syntax

__description__ = "Python app builder with a template."


def get_parser():
    parser = argparse.ArgumentParser(description=__description__)
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument('--command', help='command', nargs='?', choices=["init"], default="init")
    return parser


def main():
    banner = generate_banner(
        app_name="builder",
        author="jinwoo",
        description="Initialize Python Development Environment",
        font="graffiti",
        version=_version
    )
    print(banner)

    parser = get_parser()
    args, unknown = parser.parse_known_args()

    if args.command == "init":
        generated_file = AppGenerator(app_name="default_app").run()
        console = Console()
        print("Successful file generation")
        with open(generated_file, "rt") as code_file:
            syntax = Syntax(code_file.read(), "python")
        console.print(syntax)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pawn.console.log("[red] KeyboardInterrupt")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import time
import hashlib
import base64
from os import path
from coincurve import PrivateKey, PublicKey
from pawnlib.typing import check, date_utils, random_private_key, Namespace, fill_required_data_arguments, is_hex, sys_exit
from pawnlib.config import pawnlib_config as pawn
from pawnlib.output import is_file, is_json, open_json, check_file_overwrite, NoTraceBackException
from pawnlib.config import pawn, pconf, NestedNamespace
from pawnlib.input import PromptWithArgument, PrivateKeyValidator, StringCompareValidator, PrivateKeyOrJsonValidator
import json
import glob

from eth_keyfile import create_keyfile_json, extract_key_from_keyfile, decode_keyfile_json
from copy import deepcopy
from typing import Optional
from InquirerPy import inquirer
from InquirerPy.validator import PathValidator

compressed = False

translator = str.maketrans({
    "\\": "\\\\",
    "{": "\\{",
    "}": "\\}",
    "[": "\\[",
    "]": "\\]",
    ".": "\\."
})


def guess_wallet_type(data):
    """
    Guesses the type of wallet based on the provided data.

    :param data: The data to analyze and determine the wallet type.
    :type data: str or object

    :return: The guessed wallet type. It can be "private_key" for a private key wallet, "json" for a JSON wallet, or None if the type cannot be determined.
    :rtype: str or None

    This function attempts to determine the type of a wallet based on the provided data.
    If the `data` parameter is a string and its length is either 66 or 64 characters, it is considered a PrivateKey wallet type.
    If the `data` parameter is a valid JSON object, it is considered a JSON wallet type.
    If the wallet type cannot be determined, None is returned.

    Example:

    .. code-block:: python

        # Example 1: PrivateKey wallet type
        wallet_data = "0x1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef"
        wallet_type = guess_wallet_type(wallet_data)
        # wallet_type = "private_key"

        # Example 2: JSON wallet type
        wallet_data = '{"address": "0x1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef"}'
        wallet_type = guess_wallet_type(wallet_data)
        # wallet_type = "json"

        # Example 3: Unknown wallet type
        wallet_data = 12345
        wallet_type = guess_wallet_type(wallet_data)
        # wallet_type = None

    :raises: None
    """
    if isinstance(data, str) and len(data) == 66 or len(data) == 64:
        pawn.console.log("Wallet type is PrivateKey")
        return "private_key"
    elif is_json(data):
        pawn.console.log("Wallet type is JSON")
        return "json"
    else:
        pawn.console.log("Unknown wallet type")
        return None


class WalletCli:
    def __init__(self, args=None):

        self._args = args
        required_args = dict(
            load_type="",
            keystore="",
            base_dir="./",
            password=""
        )
        self._args = fill_required_data_arguments(required=required_args)
        self._wallet = ""

    def load(self):
        keystore_json = {}
        if getattr(self._args, "keystore", None):
            if isinstance(self._args.keystore, str):
                self._args.load_type = "text"
            else:
                self._args.load_type = "file"
            load_type = self._args.load_type
        else:
            load_type = PromptWithArgument(
                message="How to load the keystore ?",
                choices=
                [
                    {"name": "[file] From JSON file", "value": "file"},
                    {"name": "[text] From text (Copy&Paste)", "value": "text"},
                ],
                long_instruction="\nUse the up/down keys to select",
                type="list",
                max_height="60%",
                default="",
                argument="load_type",
                # verbose=True
            ).select()

        _required_password = False
        _keystore = ""
        _password = ""

        if load_type == "file":
            self._args.base_dir = inquirer.filepath(
                message="Enter the default directory where the keystore file is located:",
                default=self._args.base_dir,
                validate=PathValidator(is_dir=True, message="Input is not a directory"),
                only_directories=True,
            ).execute()

            # src_path = inquirer.filepath(
            #     message="Enter keystore file to sign:",
            #     default=self._args.base_dir,
            #     validate=PathValidator(is_file=True, message="Input is not a file"),
            #     only_files=True,
            # ).execute()

            regex_file = f"{self._args.base_dir}/*.json"
            json_file_list = glob.glob(regex_file)
            if len(json_file_list) <= 0:
                raise ValueError(f"[red] Cannot found JSON file - '{regex_file}'")

            _keystore = PromptWithArgument(
                message="Select the keystore file",
                choices=json_file_list,
                long_instruction="Use the up/down keys to select",
                instruction=f"(base_dir={self._args.base_dir})",
                type="list",
                default="",
                argument="keystore",
            ).select()
            _required_password = True

        elif load_type == "text":
            _keystore = PromptWithArgument(
                message="Input the keystore json or private key text",
                default="",
                argument="keystore",
                validate=StringCompareValidator(operator=">=", length=1),
            ).prompt()

        if _keystore:
            keystore = str(_keystore).strip()
            if is_file(keystore) and is_json(keystore):
                pawn.console.debug(f"Found Keystore JSON file - {keystore}")
                try:
                    keystore_json = open_json(keystore)
                except ValueError:
                    pawn.console.log(f"[red]Invalid JSON file - {keystore}")
                _required_password = True
            elif is_hex(keystore):
                if is_private_key(keystore):
                    pawn.console.log("Found Private key")
                    keystore_json = keystore
                    _required_password = False
                else:
                    pawn.console.log(f"[red]Invalid Private key len={len(keystore)}")
                    exit()
            else:
                try:
                    keystore_json = json.loads(keystore)
                    if isinstance(keystore_json, dict):
                        pawn.console.log("[green][OK] Loaded keystore file - JSON object")
                    else:
                        raise ValueError("Invalid JSON or Keystore text")
                    _required_password = True
                except Exception as e:
                    raise ValueError(f"[red][Error] cannot load - {e}")

        if _required_password:
            _password = PromptWithArgument(
                message="Enter password for private key",
                type="password",
                default="",
                argument="password",
                invalid_message="Requires at least one character.",
                validate=lambda result: len(result) >= 1,
            ).prompt()

        if keystore_json:
            self._wallet = load_wallet_key(keystore_json, _password)
            if self._wallet:
                self.print_wallet()
            else:
                pawn.console.log(f"[red][ERROR] Not Loaded wallet")

        return self._wallet

    def create(self, is_store_file=True):
        PromptWithArgument(
            message="Enter your private key (default: empty is random)",
            type="input",
            default="",
            argument="private_key",
            validate=PrivateKeyValidator(allow_none=True),
            # verbose=0,
        ).prompt()

        PromptWithArgument(
            message="Enter password for private key",
            type="password",
            default="",
            argument="password",
            invalid_message="Requires at least one character.",
            validate=lambda result: len(result) >= 1,
        ).prompt()

        if not self._args.private_key:
            self._args.private_key = random_private_key()

        self._wallet = load_wallet_key(self._args.private_key, password=self._args.password)
        self.print_wallet()

        if is_store_file:
            default_filename = f"{self._wallet.get('address')}_{date_utils.todaydate('ms_text')}.json"
            PromptWithArgument(
                message="Enter the name of JSON file to be saved.",
                default=default_filename,
                argument="keystore",
                invalid_message="Requires at least one character.",
                validate=lambda result: len(result) >= 1,
            ).prompt()
            if check_file_overwrite(filename=self._args.keystore):
                try:
                    wallet = generate_wallet(
                        file_path=self._args.keystore,
                        password=self._args.password,
                        overwrite=False,
                        private_key=self._args.private_key,
                        expected_address=self._wallet.get('address'),
                    )
                    pawn.console.log(f"Generate Wallet - {wallet.get_hx_address()} to '{self._args.keystore}'")
                except Exception as e:
                    pawn.console.log(f"[red][ERROR] Generate wallet - {e}")

    def print_wallet(self):
        if self._wallet:
            pawn.console.print(f"\n"
                               f" 🔑 address={self._wallet.get('address')}\n"
                               f" 🔑 public_key={self._wallet.get('public_key')}")
            if pawn.get('PAWN_DEBUG'):
                pawn.console.print(f" 🔑 public_key_long={self._wallet.get('public_key_long')}")
                pawn.console.print(f" 🔑 private_key={self._wallet.get('private_key')}")
            pawn.console.print("")


def store_keystore_file_on_the_path(file_path, json_string, overwrite=False):
    """Stores a created keystore string data which is JSON format on the file path.
    :param file_path: The path where the file will be saved. type(str)
    :param json_string: Contents of the keystore.
    """
    if not overwrite and path.isfile(file_path):
        raise FileExistsError

    with open(file_path, 'wt') as f:
        f.write(json_string)


def generate_wallet(file_path=None, password=None, overwrite=False, private_key=None, expected_address=None):
    singer = IcxSigner(data=private_key)
    if not file_path:
        file_path = f"{singer.get_hx_address()}_{date_utils.todaydate('ms_text')}.json"
    singer.store(file_path, password, overwrite, expected_address=expected_address)
    return singer


def _parse_keystore_key(file=None, password=None, private_key_hex=None, use_namespace=False):
    if private_key_hex:
        if private_key_hex.startswith("0x"):
            private_key_hex = private_key_hex[2:]
        private_key = bytes.fromhex(private_key_hex)
    else:
        try:
            if not password:
                raise ValueError(f"Invalid password -> '{password}'")
            private_key: bytes = decode_keyfile_json(file, bytes(password, 'utf-8'))
        except ValueError as e:
            if "MAC mismatch" in str(e):
                e = "Wrong password"
            raise ValueError(e)

    _private_key = PrivateKey(private_key)
    _public_key_long: bytes = _private_key.public_key.format(compressed=False)
    _public_key: bytes = _private_key.public_key.format(compressed=True)
    address = f"hx{get_address(pubkey_bytes=_public_key_long).hex()}"
    wallet_dict = {
        # "private_key": "0x" + private_key.hex(),
        "private_key": private_key.hex(),
        "address": address,
        "public_key": _public_key.hex(),
        "public_key_long": _public_key_long.hex()
    }
    pawn.console.debug(wallet_dict)
    if use_namespace:
        return NestedNamespace(**wallet_dict)
    return wallet_dict


def is_private_key(private_key):
    if isinstance(private_key, str):
        if (len(private_key) == 64 or len(private_key) == 66) and check.is_hex(private_key):
            return True
    return False


def exit_on_failure(raise_on_failure, exception):
    if raise_on_failure:
        raise NoTraceBackException(exception)
    else:
        pawn.console.log(f"[red][ERROR][/red] {exception}")


def load_wallet_key(file_or_object=None, password=None, raise_on_failure=True, use_namespace=False):
    if isinstance(password, (dict, list, tuple)):
        raise ValueError(f"Wrong password type => {password} ({type(password)})")

    _keystore_params = dict()

    if isinstance(file_or_object, dict):
        pawn.console.debug("Loading wallet from keystore file - JSON dict")
        _keystore_params = dict(
            file=file_or_object,
            password=password,
        )
    elif is_file(file_or_object):
        try:
            pawn.console.debug(f"Loading wallet from keystore file : {file_or_object}")
            _keystore_params = dict(
                file=open_json(file_or_object),
                password=password,
            )
        except ValueError as e:
            pawn.console.log(f"[bold red] Open File - {e}")
    elif is_private_key(file_or_object):
        pawn.console.debug("Loading wallet from a Private Key")
        _keystore_params = dict(
            private_key_hex=file_or_object,
        )
    else:
        pawn.console.debug("Loading wallet from JSON data")
        try:
            file_json = json.loads(file_or_object)
            _keystore_params = dict(
                file=file_json,
                password=password,
            )
        except Exception as e:
            pawn.console.log(f"[bold red] Failed to load JSON data - {e}")

    if _keystore_params:
        try:
            return _parse_keystore_key(use_namespace=use_namespace, **_keystore_params)
        except Exception as e:
            exit_on_failure(raise_on_failure=raise_on_failure, exception=e)

    return {}


def generate_keys():
    """generate privkey and pubkey pair using coincurve.

    Returns:
        tuple: privkey(bytes, 32), pubkey(bytes, 65)
    """
    privkey = PrivateKey()
    privkey_bytes = privkey.secret
    pubkey_bytes = privkey.public_key.format(compressed=False)

    return privkey_bytes, pubkey_bytes


def get_address(pubkey_bytes):
    """generate address from public key.

    Args:
        pubkey_bytes(bytes): public key bytes

    Returns:
        bytes: icx address (20bytes)
    """

    # Remove the first byte(0x04) of pubkey
    return hashlib.sha3_256(pubkey_bytes[1:]).digest()[-20:]


def verify_recoverable_signature(msg_hash, signature_bytes, recovery_id):
    """
    Args:
        msg_hash(bytes): 256bit hash value
        signature_bytes(bytes):
        recovery_id(int):

    Returns:
    """
    pubkey, signature = \
        recover_signature(msg_hash, signature_bytes, recovery_id)

    return pubkey.ecdsa_verify(msg_hash, signature, True)


def recover_signature(msg_hash, signature_bytes, recovery_id):
    """
    Args:
        msg_hash(bytes): sha3 256bit hash value
        signature_bytes(bytes):
        recovery_id(int):

    Returns:
        pubkey(PublicKey):
        signature(bytes):
    """
    recoverable_signature = signature_bytes + bytes([recovery_id])
    public_key = PublicKey.from_signature_and_message(recoverable_signature, msg_hash, hasher=None)

    signature = PublicKey.from_signature_and_message(recoverable_signature, msg_hash, hasher=None).format()

    return public_key, signature


class IcxSigner(object):
    """Digital Signing  using coincurve
    """

    def __init__(self, data=None, raw=True):
        """Constructor

        Args:
            data(object): bytes or der
            raw(bool): True(bytes) False(der)
        """
        self._private_key_hex = None
        self._private_key_bytes = None
        if data:
            self._check_private_key(data)

        self.__privkey = PrivateKey(self._private_key_bytes)

    def _check_private_key(self, private_key=None):

        if isinstance(private_key, bytes):
            self._private_key_hex = private_key.hex()
        elif check.is_hex(private_key):
            self._private_key_hex = private_key
        else:
            raise ValueError(f"Invalid Private Key - {private_key}")

        if self._private_key_hex.startswith("0x"):
            self._private_key_hex = self._private_key_hex[2:]

        self._private_key_bytes = bytes.fromhex(self._private_key_hex)
        # pawn.console.debug(f"[green] {self._private_key_bytes}")

    def set_privkey_bytes(self, data):
        """Set private key using private key data in bytes.

        Args:
            data(bytes): private key data
        """
        self.__privkey = PrivateKey(data)

    def get_privkey_bytes(self):
        """Get private key data in bytes.

        Returns:
            bytes: private key data (32 bytes)
        """
        return self.__privkey.secret

    def get_pubkey_bytes(self):
        return self.__privkey.public_key.format(compressed=compressed)

    def get_address(self) -> bytes:
        """Create an address with pubkey.
        address is made from pubkey.

        Returns:
            str: address represented in hexadecimal string starting with '0x'
        """
        pubkey_bytes = self.get_pubkey_bytes()
        return get_address(pubkey_bytes)

    def get_hx_address(self):
        """Create an address with pubkey.
        address is made from pubkey.

        Returns:
            str: address represented in hexadecimal string starting with '0x'
        """

        return f"hx{self.get_address().hex()}"

    def sign_tx(self, tx=None):
        if isinstance(tx, dict) and tx.get('params'):
            tx_hash_bytes = get_tx_hash(params=tx['params'])
            signature_bytes = self.sign_recoverable(tx_hash_bytes)
            signature = signature_bytes[:64]
            recovery_id = signature_bytes[64:]
            signature_bytes_big = signature + recovery_id
            tx['params']['signature'] = base64.b64encode(signature_bytes_big).decode()
        return tx

    def sign(self, msg_hash):
        """Make a signature using the hash value of msg.

        Args:
            msg_hash(bytes): msg_hash = sha3_256(msg)

        Returns:
            bytes: signature bytes
        """
        privkey = self.__privkey
        signature = privkey.sign(msg_hash, hasher=None)
        return signature

    def store(self, file_path: str, password: str, overwrite: bool = False, expected_address: str = None):
        try:
            key_store_contents = create_keyfile_json(
                self.get_privkey_bytes(),
                bytes(password, 'utf-8'),
                iterations=16384,
                kdf="scrypt"
            )
            key_store_contents['address'] = self.get_hx_address()
            key_store_contents['coinType'] = 'icx'

            # validate the  contents of a keystore file.
            if expected_address and expected_address != self.get_hx_address():
                raise ValueError(f"Not expected address => expected({expected_address}) != real({self.get_hx_address()})")

            import json
            if key_store_contents:
                json_string_keystore_data = json.dumps(key_store_contents)
                store_keystore_file_on_the_path(file_path, json_string_keystore_data, overwrite)
                pawn.console.debug(f"Stored Wallet. Address: {self.get_hx_address()}, File path: {file_path}")
        except FileExistsError:
            raise ValueError("File already exists.")
        except PermissionError:
            raise ValueError("Not enough permission.")
        except FileNotFoundError:
            raise ValueError("File not found.")
        except IsADirectoryError:
            raise ValueError("Directory is invalid.")

    def sign_recoverable(self, msg_hash):
        """Make a recoverable signature using message hash data
        We can extract public key from recoverable signature.

        Args:
            msg_hash(bytes): hash data of message

        Returns:
            tuple:
                bytes: 65 bytes data
                int: recovery id
        """
        privkey = self.__privkey
        signature_bytes = privkey.sign_recoverable(msg_hash, hasher=None)
        return signature_bytes
    @staticmethod
    def from_bytes(data):
        return IcxSigner(data, raw=True)

    @staticmethod
    def from_der(data):
        return IcxSigner(data, raw=False)


class IcxSignVerifier(object):
    """Digital signature verification
    """

    def __init__(self, data):
        """
        Args:
            data(bytes): 65 bytes data which PublicKey.serialize() returns

        Returns:
            None
        """
        self.__pubkey = PublicKey(data)

    def get_address(self):
        """Create an address with pubkey.
        address is made from pubkey.

        Returns:
            str: address represented in hexadecimal string starting with '0x'
        """
        pubkey_bytes = self.__pubkey.format(compressed=compressed)
        return get_address(pubkey_bytes)

    def verify(self, msg_hash, signature_bytes):
        """Check whether signature is valid or not.

        Args:
            pubkey_bytes(bytes): byte data of pubkey
            msg_hash(bytes): hash value of msg
            signature_bytes(bytes): signature data

        Returns:
            bool: the result of signature verification
        """
        return self.__pubkey.verify(signature_bytes, msg_hash)

    @staticmethod
    def from_bytes(data):
        """
        Args:
            data(bytes): bytes data which PublicKey.serialize() returns

        Returns:
            None
        """
        return IcxSignVerifier(data)


def get_timestamp_us():
    """Get epoch time in us.
    """
    return int(time.time() * 10 ** 6)


def icx_to_wei(icx):
    """Convert amount in icx unitt to wei unit.

    Args:
        icx(float): float value in icx unit

    Returns:
        int: int value in wei unit
    """
    return int(icx * 10 ** 18)


def get_string_decimal(value, place):
    """value를 10의 place 제곱으로 나눈 값을 string으로 변환하여 반환

    Args:
        value(int)
        place : 10의 몇 제곱을 나눌지 입력받음
    """
    str_value = str(value)
    if value >= 10 ** place:
        str_int = str_value[:-place]
        str_decimal = str_value[-place:]
        result = f'{str_int}.{str_decimal}'
        return result
    else:
        zero_prefix = "0."
        value_digits = len(str_value)  # value_digits : 몇자릿수인지 계산
        zero_padding = "0" * (place - value_digits)
        result = f'{zero_prefix}{zero_padding}{value}'
        return result


def sha3_256(data):
    """Get hash value using sha3_256 hash function

    Args:
        data(bytes): data to hash

    Returns:
        bytes: 256bit hash value (32 bytes)
    """
    return hashlib.sha3_256(data).digest()

#
def get_tx_hash(params=None):
    """Create tx_hash from params object.

    Args:
        params(dict): the value of 'params' key in jsonrpc

    Returns:
        bytes: sha3_256 hash value
        :param params:
    """
    tx_phrase = serialize(params)
    pawn.console.debug(f"serialize tx={tx_phrase}")
    return sha3_256(tx_phrase)


def get_tx_phrase(method, params):
    """Create tx phrase from method and params.
    tx_phrase means input text to create tx_hash.

    Args:
        params(dict): the value of 'params' key in jsonrpc

    Returns:
        str: sha3_256 hash format without '0x' prefix
    """
    keys = [key for key in params]
    keys.sort()

    key_count = len(keys)
    if key_count == 0:
        return method

    phrase = f'{keys[0]}.{params[keys[0]]}'
    for i in range(1, key_count):
        key = keys[i]
        phrase += f'.{key}.{params[key]}'

    return f'{method}.{phrase}'


def sign_recoverable(privkey_bytes, tx_hash_bytes):
    """
    Args:
        tx_hash(bytes): 32byte tx_hash data

    Returns:
        bytes: signature_bytes + recovery_id(1)
    """
    signer = IcxSigner.from_bytes(privkey_bytes)
    signature_bytes, recovery_id = signer.sign_recoverable(tx_hash_bytes)

    # append recover_id(1 byte) to signature_bytes.
    return bytes(bytearray(signature_bytes) + recovery_id.to_bytes(1, 'big'))


def __make_params_serialized(json_data: dict) -> str:

    def encode(data) -> str:
        if isinstance(data, dict):
            return encode_dict(data)
        elif isinstance(data, list):
            return encode_list(data)
        else:
            return escape(data)

    def encode_dict(data: dict) -> str:
        result = ".".join(_encode_dict(data))
        return "{" + result + "}"

    def _encode_dict(data: dict) -> list:
        for key in sorted(data.keys()):
            yield key
            yield encode(data[key])

    def encode_list(data: list) -> str:
        result = ".".join(_encode_list(data))
        return f"[" + result + "]"

    def _encode_list(data: list) -> list:
        for item in data:
            yield encode(item)

    def escape(data) -> str:
        if data is None:
            return "\\0"

        data = str(data)
        return data.translate(translator)

    return ".".join(_encode_dict(json_data))


def serialize(params: dict) -> bytes:
    """
    Serialized params of an original JSON request starting with `icx_sendTransaction`
    to generate a message hash for a signature.
    :param params: params in a original JSON request for transaction.
    :return: serialized params.
    For example, data like `icx_sendTransaction.<key1>.<value1>.<key2>.<value2>` is converted to bytes.
    """
    copy_tx = deepcopy(params)
    key_name_for_tx_hash = __get_key_name_for_tx_hash(params)

    if key_name_for_tx_hash in copy_tx:
        del copy_tx[key_name_for_tx_hash]

    if 'signature' in copy_tx:
        del copy_tx['signature']

    partial_serialized_params = __make_params_serialized(copy_tx)
    return f"icx_sendTransaction.{partial_serialized_params}".encode()


def generate_message(params: dict) -> str:
    """
    Generates transaction's message hash from params in request for transaction.
    :param params: params in request for transaction.
    :return: the 256 bit hash digest of a message. Hexadecimal encoded.
    """
    bytes_message_hash = serialize(params)
    return sha3_256(bytes_message_hash).hexdigest()


def __get_key_name_for_tx_hash(params: dict) -> Optional[str]:
    if __get_tx_version(params) == hex(2):
        return "tx_hash"
    else:
        return None


def __get_tx_version(params: dict) -> str:
    if 'version' not in params:
        return hex(2)
    else:
        return params['version']



import os
from pathlib import Path
from typing import Optional
import json
import logging


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return getinstance


@singleton
class Configure:
    def __init__(self, file_name="../config.ini", path=None, section=None):

        self.file_name = file_name
        if path:
            self.path = path
        else:
            self.path = os.path.dirname(os.path.abspath(__file__))
        self.full_path = os.path.join(self.path, self.file_name)
        self.section = section
        self.name = "default"

    def _get_config(self, ):
        self.config.optionxform = str  # change to uppercase
        self.config.read(self.full_path)
        return self._converter(self.config._sections)


class Config:
    CONFIG_DEFAULT_FILE = 'configure.json'

    def __init__(self, config_path: Optional[Path] = None, logger=None):
        super().__init__()
        self._current_path: Optional[Path] = None
        self._config_path = config_path or self.get_path(self.CONFIG_DEFAULT_FILE)
        self._config = {}
        self._logger = logger
        self._log_options = None
        self._load_config_file(self._config_path, check_exist=False)
        # self._set_common_default()
        self._load_config_env()
        # self._update_log_options()

    def load_config(self, config_path: Path):
        """Load User configuration file.
        :param config_path: The str path created with pathlib.Path is recommended.
        """
        self._load_config_file(config_path=config_path)
        self._update_log_options()

    def __setattr__(self, key, value):
        if key in ['_current_path', '_config_path', '_config', '_logger', '_log_options']:
            self.__dict__[key] = value
        else:
            self._config[key] = value

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__getattribute__(name)
        return self[name]

    def __setitem__(self, key, item):
        self._config[key] = item
        # self._update_log_options()

    def __getitem__(self, key):
        return self._config[key]

    def _set_common_default(self):
        if 'debug' not in self._config:
            self._logger.debug("config['debug'] is not configured. So default(False) is used.")
            self._config['debug'] = False
        if 'log_level' not in self._config:
            self._logger.debug("config['log_level'] is not configured. So default(logging.WARNING) is used.")
            self._config['log_level'] = logging.getLevelName(logging.WARNING)
        if 'log_format' not in self._config:
            self._logger.debug("config['log_format'] is not in configured. So default is used.")
            self._config['log_format'] = \
                "%(asctime)s,%(msecs)03d %(process)d %(thread)d %(levelname)s %(filename)s(%(lineno)d) %(message)s"

    def _load_config_file(self, config_path: Path, check_exist=True):
        if not check_exist and not config_path.is_file():
            return

        with open(config_path, 'r') as config_file:
            self._config.update(json.load(config_file))
            self._current_path = config_path.parent

    def _load_config_env(self):
        for config_name in self._config:
            env_value = os.getenv(config_name) or os.getenv(str.upper(config_name))
            if env_value is not None:
                if isinstance(self[config_name], (bool, int, float)):
                    env_value = eval(env_value)
                self[config_name] = env_value

    def write_config(self, config_path: Optional[Path] = None, encoding: str = 'utf-8'):
        """You can set config values in code, and save it as a file.

        :param config_path: Path or CONFIG_DEFAULT_FILE.
        :param encoding: file encoding.

        :return:

        """
        write_path = config_path or self._config_path
        with open(write_path, 'w', encoding=encoding) as config_file:
            json.dump(self._config, config_file, indent=4, sort_keys=True)

    def get_path(self, path: str) -> Path:
        """Get Path from the directory where the configure.json file is.
        :param path: file_name or path
        :return:
        """
        root_path = self._current_path or Path(os.path.join(os.getcwd()))
        return root_path.joinpath(path)


if __name__ == '__main__':
    CFG = Configure()





#!/usr/bin/env python3
import os
import logging
import sys
from logging import handlers
import traceback
import datetime
from pawnlib.config.globalconfig import pawnlib_config
from rich.logging import RichHandler
from typing import Callable

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class CustomLog:
    """CustomLog

    :param name: logger name

    Example:

        .. code-block:: python

            from pawnlib.utils.log import CustomLog

            file_name = './time_log.txt'
            logger = CustomLog("custom_log")
            logger.set_level('DEBUG')
            logger.stream_handler("INFO")
            logger.time_rotate_handler(filename=file_name,
                                       when="M",
                                       interval=2,
                                       backup_count=3,
                                       level="INFO"
                                       )
            idx = 1
            logger.log.debug(logger.log_formatter(f'debug {idx}'))
            logger.log.info(logger.log_formatter(f'info {idx}'))
            logger.log.warning(logger.log_formatter(f'warning {idx}'))
            logger.log.error(logger.log_formatter(f'error {idx}'))
            logger.log.critical(logger.log_formatter(f'critical {idx}'))

    """

    def __init__(self, name):
        self.log = logging.getLogger(name)
        self.log.propagate = True
        # self.formatter = logging.Formatter("%(levelname).1s|%(asctime)s.%(msecs)06d|-|%(name)s|%(message)s", "%Y%m%d-%H:%M:%S")
        # self.formatter = logging.Formatter(f"%(levelname).1s|%(asctime)s.%(msecs)06d|-|%(name)s|%(filename)s:%(lineno)d %(funcName)-15s| %(message)s", "%Y%m%d-%H:%M:%S")
        self.formatter = logging.Formatter(
            f"%(levelname).1s|%(asctime)s.%(msecs)06d|-|%(name)s|%(filename)s:%(lineno)d| %(message)s",
            "%Y%m%d-%H:%M:%S"
        )
        self.levels = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "WARNING": logging.WARNING,
            "ERROR": logging.ERROR,
            "CRITICAL": logging.CRITICAL
        }

    def set_level(self, level):
        self.log.setLevel(self.levels[level])

    def log_formatter(self, msg):
        """

        :param msg:
        :return:
        """
        log_str = f"{msg}"
        return log_str

    def stream_handler(self, level):
        """
        :param level:

        Note:
            level

            * "DEBUG" : logging.DEBUG ,
            * "INFO" : logging.INFO ,
            * "WARNING" : logging.WARNING ,
            * "ERROR" : logging.ERROR ,
            * "CRITICAL" : logging.CRITICAL ,
        :return:
        """
        _stream_handler = logging.StreamHandler()
        _stream_handler.setLevel(self.levels[level])
        _stream_handler.setFormatter(self.formatter)
        self.log.addHandler(_stream_handler)
        return self.log

    def file_handler(self, file_name, mode):
        """

        :param file_name: ~.txt / ~.log
        :param mode: "w" / "a"
        :return:
        """
        _file_handler = logging.FileHandler(file_name, mode=mode)
        _file_handler.setLevel(logging.DEBUG)
        _file_handler.setFormatter(self.formatter)
        self.log.addHandler(_file_handler)
        return self.log

    def file_rotating_handler(self, file_name, mode, level, backup_count, log_max_size):
        """

        :param file_name: file의 이름 , ~.txt / ~.log
        :param mode: "w" / "a"
        :param backup_count: backup할 파일 개수
        :param log_max_size: 한 파일당 용량 최대
        :param level:

        > "DEBUG" : logging.DEBUG ,
        > "INFO" : logging.INFO ,
        > "WARNING" : logging.WARNING ,
        > "ERROR" : logging.ERROR ,
        > "CRITICAL" : logging.CRITICAL ,
        :return:
        """

        _file_handler = logging.handlers.RotatingFileHandler(
            filename=file_name,
            maxBytes=log_max_size,
            backupCount=backup_count,
            mode=mode)
        _file_handler.setLevel(self.levels[level])
        _file_handler.setFormatter(self.formatter)
        self.log.addHandler(_file_handler)
        return self.log

    def time_rotate_handler(self,
                            filename='./log.txt',
                            when="M",
                            level="DEBUG",
                            backup_count=4,
                            atTime=datetime.time(0, 0, 0),
                            interval=1):
        """
        :param level:
        :param filename:
        :param when: 저장 주기
        :param interval: 저장 주기에서 어떤 간격으로 저장할지
        :param backup_count: 5
        :param atTime: datetime.time(0, 0, 0)
        :return:
        """
        _file_handler = logging.handlers.TimedRotatingFileHandler(
            filename=filename,
            when=when,  # W0
            backupCount=backup_count,
            interval=interval,
            atTime=atTime)
        _file_handler.setLevel(self.levels[level])
        _file_handler.setFormatter(self.formatter)
        self.log.addHandler(_file_handler)
        return self.log


class AppLogger:
    """

    AppLogger

    :param app_name: application name(=file name)
    :param log_level: log level
    :param log_path: log file path
    :param stdout: Enable stdout, Adding Hook for another library logging.
    :param markup: Enable markup for stdout logging.
    :param stdout_level: stdout log level
    :param stdout_log_formatter: stdout log formatter (function)
    :param log_format: log format / [%(asctime)s] %(name)s::" "%(filename)s/%(funcName)s(%(lineno)d) %(message)s
    :param use_hook_exception: Select whether to log exception errors.
    :param exception_handler: Exception handling function
    :param debug:

    Example:

        .. code-block:: python

            from pawnlib.utils import log

            app_logger, error_logger = log.AppLogger().get_logger()
            app_logger.info("This is a info message")
            error_logger.error("This is a info message")


    Example2:

        .. code-block:: python

            from pawnlib.config.globalconfig import pawnlib_config as pawn

            app_logger, error_logger = log.AppLogger(
                app_name="app",
                log_path="./logs",
                stdout=True
            ).set_global()

            pawn.app_logger.info("This is a info message")
            pawn.error_logger.error("This is a error message")

            # >>>
            [2022-07-25 18:52:44,415] INFO::app_logging_test.py/main(38) This is a info message
            [2022-07-25 18:52:44,416] ERROR::app_logging_test.py/main(39) This is a info message


    """
    _logger = None

    def __init__(self,
                 app_name: str = "default",
                 log_level: Literal["INFO", "WARN", "DEBUG"] = "INFO",
                 log_path: str = "./logs",
                 markup: bool = True,
                 stdout: bool = False,
                 stdout_level: Literal["INFO", "WARN", "DEBUG", "NOTSET"] = "INFO",
                 stdout_log_formatter: Callable = "%H:%M:%S,%f",
                 log_format: str = None,
                 debug: bool = False,
                 use_hook_exception: bool = True,
                 exception_handler: Callable = "",
                 **kwargs
                 ):
        self.app_name = app_name
        self.log_path = log_path
        self.debug = debug
        self.stdout = stdout
        self.stdout_level = stdout_level
        self.stdout_log_formatter = stdout_log_formatter
        self.markup = markup
        self.log_level = log_level
        self.use_hook_exception = use_hook_exception
        self.kwargs = kwargs

        if self.use_hook_exception:
            if exception_handler:
                sys.excepthook = exception_handler
            else:
                sys.excepthook = self.handle_exception

        if log_format:
            self.log_format = log_format
        else:
            self.log_format = "[%(asctime)s] %(name)s::" "%(filename)s/%(funcName)s(%(lineno)d) %(message)s"

        self.log_formatter = logging.Formatter(self.log_format)

        self._logger = self.set_logger(self.log_level)
        self._error_logger = self.set_logger("ERROR")

    def get_realpath(self):
        path = os.path.dirname(os.path.abspath(__file__))
        parent_path = os.path.abspath(os.path.join(path, ".."))
        return parent_path

    def set_logger(self, log_type="INFO"):
        # log_path = f"{self.get_realpath()}/logs"
        # print(f"log_path={self.log_path}")
        if not os.path.isdir(self.log_path):
            os.mkdir(self.log_path)

        _logger = logging.getLogger(log_type)
        stack = traceback.extract_stack()
        _logger.setLevel(getattr(logging, log_type))

        if log_type == "ERROR":
            filename = f"{self.app_name}.{str(log_type).lower()}.log"
        else:
            filename = f"{self.app_name}.log"

        logfile_filename = "%s/%s" % (self.log_path, filename)

        file_handler = self.time_rotate_handler(
            filename=logfile_filename,
            when='midnight',
            interval=1,
            encoding='utf-8',
            backup_count=10
        )
        file_handler.suffix = '%Y%m%d'
        file_handler.setFormatter(self.log_formatter)
        _logger.addHandler(file_handler)

        if self.stdout:
            from rich.text import Text
            if self.stdout_log_formatter:
                log_time_formatter = lambda dt: Text.from_markup(f"[{dt.strftime(self.stdout_log_formatter)[:-3]}]")
            else:
                log_time_formatter = None
            # if self.stdout_log_formatter:
            #
            # else:
            #     log_time_formatter = lambda dt: Text.from_markup(f"[{dt.strftime('%H:%M:%S,%f')[:-3]}]")

            logging.basicConfig(
                # level=self.stdout_level, format="%(message)s", datefmt="[%Y-%m-%d %H:%M:%S.%f]", handlers=[RichHandler(rich_tracebacks=True)]
                level=self.stdout_level,
                format="%(message)s",
                handlers=[
                    RichHandler(
                        rich_tracebacks=True,
                        log_time_format=log_time_formatter,
                        markup=self.markup,
                        **self.kwargs
                        # show_path=False,
                        # log_time_format=lambda dt: f"[{dt.strftime('%Y-%m-%d %H:%M:%S,%f')[:-3]}]",
                        # log_time_format=lambda dt:Text.from_markup(f"[red]{dt.ctime()}")
                    )
                ]
            )
        # _logger.addHandler(self.add_stream_handler(level=log_type))
        return _logger

    def add_stream_handler(self, level):
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(level)
        stream_handler.setFormatter(self.log_formatter)
        return stream_handler

    def time_rotate_handler(self,
                            filename='./log.txt',
                            when="M",
                            backup_count=4,
                            atTime=datetime.time(0, 0, 0),
                            interval=1,
                            encoding="utf-8"
                            ):
        file_handler = logging.handlers.TimedRotatingFileHandler(
            filename=filename,
            when=when,  # W0
            backupCount=backup_count,
            interval=interval,
            atTime=atTime,
            encoding=encoding
        )
        return file_handler

    def get_logger(self):
        """
        Get the logger

        :return:
        """
        return self._logger, self._error_logger

    def set_global(self):
        """
        Add global config in pawnlib

        :return:
        """
        pawnlib_config.set(
            PAWN_APP_LOGGER=self._logger,
            PAWN_ERROR_LOGGER=self._error_logger,
        )

    def handle_exception(self, exc_type, exc_value, exc_traceback):
        if self.use_hook_exception and self._error_logger:
            self._error_logger.error("Unexpected exception", exc_info=(exc_type, exc_value, exc_traceback))



import re

from pawnlib.config.globalconfig import pawnlib_config as pawn
import socket
import time
import asyncio
import requests
from concurrent.futures import ThreadPoolExecutor
from timeit import default_timer
from pawnlib.utils import http
from pawnlib.typing import is_valid_ipv4, todaydate, shorten_text
from pawnlib.output import PrintRichTable


try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

prev_getaddrinfo = socket.getaddrinfo


class OverrideDNS:
    """

    Change the Domain Name using socket

    Example:

        .. code-block:: python

            from pawnlib.resource import net
            net.OverrideDNS(domain=domain, ipaddr=ipaddr).set()

    """
    _dns_cache = {}

    def __init__(self, domain="", ipaddr="", port=80):
        self._dns_cache[domain] = ipaddr
        self.prv_getaddrinfo = prev_getaddrinfo

    def new_getaddrinfo(self, *args):
        if args[0] in self._dns_cache:
            if pawn.verbose:
                print("Forcing FQDN: {} to IP: {}".format(args[0], self._dns_cache[args[0]]))
            return self.prv_getaddrinfo(self._dns_cache[args[0]], *args[1:])
        else:
            return self.prv_getaddrinfo(*args)

    def set(self):
        socket.getaddrinfo = self.new_getaddrinfo

    def unset(self):
        socket.getaddrinfo = self.prv_getaddrinfo


def get_public_ip():
    """
    The get_public_ip function returns the public IP address of the machine it is called on.


    :return: The public ip address of the server

    Example:

        .. code-block:: python

            from pawnlib.resource import net
            net.get_public_ip()


    """
    try:
        public_ip = http.jequest("http://checkip.amazonaws.com", timeout=2).get('text', "").strip()
        if is_valid_ipv4(public_ip):
            return public_ip
        else:
            pawn.error_logger.error(f"An error occurred while fetching Public IP address. Invalid IPv4 address - '{public_ip}'")
            pawn.console.debug(f"An error occurred while fetching Public IP address. Invalid IPv4 address - '{public_ip}'")

    except Exception as e:
        pawn.error_logger.error(f"An error occurred while fetching Public IP address - {e}")
        pawn.console.debug(f"An error occurred while fetching Public IP address - {e}")

    return ""


class FindFastestRegion:
    def __init__(self, verbose=True, aws_regions=None):
        self.results = []
        self.verbose = verbose
        if aws_regions:
            self.aws_regions = aws_regions
        else:
            self.aws_regions = {
                "Seoul": "ap-northeast-2",
                "Tokyo": "ap-northeast-1",
                "Virginia": "us-east-1",
                "Hongkong": "ap-east-1",
                "Singapore": "ap-southeast-1",
                "Mumbai": "ap-south-1",
                "Frankfurt": "eu-central-1",
                "Ohio": "us-east-2",
                "California": "us-west-1",
                "US-West": "us-west-2",
                "Ceentral":"ca-central-1",
                "Ireland": "eu-west-1",
                "London": "eu-west-2",
                "Sydney": "ap-southeast-2",
                "São Paulo": "sa-east-1",
                "Beijing": "cn-north-1",
            }

    def run(self):
        self.results = []
        loop = asyncio.get_event_loop()
        future = asyncio.ensure_future(self.find_fastest_region())
        loop.run_until_complete(future)
        self.sorted_results()
        return self.results

    async def find_fastest_region(self):
        tasks = []
        with ThreadPoolExecutor(max_workers=10) as executor:
            loop = asyncio.get_event_loop()
            for region_name, region_code in self.aws_regions.items():
                url = f'https://s3.{region_code}.amazonaws.com/ping?x=%s' % todaydate("ms")
                tasks.append(loop.run_in_executor(executor, self.get_time, *(url, region_name)))
            await asyncio.gather(*tasks)

    def get_time(self, url, name="NULL"):
        start_time = default_timer()
        try:
            response = requests.get(f'{url}', timeout=3)
            response_text = response.text
            response_time = round(response.elapsed.total_seconds(), 3)
            status_code = response.status_code
        except:
            response_time = None
            response_text = None
            status_code = 999
        elapsed = round(default_timer() - start_time, 3)

        data = {
            "region": name,
            "time": response_time,
            "run_time": elapsed,
            "url": shorten_text(url, 50),
            # "text": response_text,
            "status_code": status_code
        }
        if data.get('time') and data.get("run_time") and data.get("status_code") == 200:
            self.results.append(data)
            if self.verbose:
                print(data)
        return data

    def sorted_results(self, key="run_time"):
        self.results = sorted(self.results, key=(lambda x: x.get(key)), reverse=False)

    def print_results(self):
        PrintRichTable(title="fast_region", data=self.results)
        pawn.console.log(f"Fastest Region={self.results[0]['region']}, time={self.results[0]['run_time']} sec")


def get_local_ip():
    """

    Get the local IP address

    :return:

    Example:

        .. code-block:: python

            from pawnlib.resource import net
            net.get_local_ip()

    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('10.255.255.255', 1))
        ipaddr = s.getsockname()[0]
    except Exception:
        ipaddr = '127.0.0.1'
    finally:
        s.close()

    if is_valid_ipv4(ipaddr):
        return ipaddr
    else:
        pawn.error_logger.error("An error occurred while fetching Local IP address. Invalid IPv4 address")
        pawn.console.debug("An error occurred while fetching Local IP address. Invalid IPv4 address")
    return ""


def get_hostname():
    """

    Get the local hostname

    :return:

    Example:

        .. code-block:: python

            from pawnlib.resource import net
            net.get_hostname()

    """
    return socket.gethostname()


def extract_host_port(host):
    """
    The extract_host_port function extracts the host and port from a string.

    :param host: Extract the hostname from the url
    :return: A tuple of the host and port number

    Example:

    .. code-block:: python

        from pawnlib.resource import net
        net.extract_host_port("http://127.0.0.1:8000")

    """
    http_regex = '^((?P<proto>https?)(://))?(?P<host>[^:/ ]+).?(?P<port>[0-9]*).*'

    regex_res = re.search(http_regex, host)
    port = 0
    if regex_res:
        if regex_res.group('port'):
            port = int(regex_res.group('port'))
        else:
            if regex_res.group('proto'):
                if regex_res.group('proto') == "http":
                    port = 80
                elif regex_res.group('proto') == "https":
                    port = 443
            else:
                port = 80
        host = regex_res.group('host')
        pawn.console.debug(f"[Regex] host={host}, port={port}, {regex_res.groupdict()}")

    return host, port


def check_port(host: str = "", port: int = 0, timeout: float = 3.0, protocol: Literal["tcp", "udp"] = "tcp") -> bool:
    """
    Returns boolean with checks if the port is open

    :param host: ipaddress os hostname
    :param port: destination port number
    :param timeout: timeout sec
    :param protocol: type of protocol
    :return: boolean

    Example:

        .. code-block:: python

            from pawnlib.resource import net
            net.check_port()

    """

    if not port:
        host, port = extract_host_port(host)
    else:
        pawn.console.debug(f"[red][Not Matched] host={host}, port={port}")

    pawn.console.debug(f"host={host}, port={port}, protocol={protocol}")

    if protocol == "tcp":
        socket_protocol = socket.SOCK_STREAM
    elif protocol == "udp":
        socket_protocol = socket.SOCK_DGRAM
    else:
        raise ValueError("Invalid socket type argument, tcp or udp")

    if host == "" or port == 0:
        raise ValueError(f"Invalid host or port, inputs: host={host}, port={port}")

    if timeout:
        socket.setdefaulttimeout(float(timeout))  # seconds (float)

    with socket.socket(socket.AF_INET, socket_protocol) as sock:
        host = http.remove_http(host)
        try:
            result = sock.connect_ex((host, port))
        except Exception as e:
            pawn.error_logger.error(f"[FAIL] {e}")
            return False

    if result == 0:
        pawn.console.debug(f"[OK] Opened port -> {host}:{port}")
        return True
    else:
        pawn.error_logger.error(f"[FAIL] Closed port -> {host}:{port}")
    return False


def listen_socket(host, port):
    """
    Create a socket object and bind it to the host and port provided.
    Listen for incoming connections on that socket, with a maximum of 5 connections in the queue.

    :param host: str - hostname of the machine where the server is running
    :param port: int - port number that the server will listen on
    :return: socket - a socket object

    Example:
        .. code-block:: python

            # create a socket object and bind it to localhost and port 8080
            sock = listen_socket("localhost", 8080)

            # listen for incoming connections
            conn, addr = sock.accept()

    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind((host, port))
    sock.listen(5)
    return sock


def wait_for_port_open(host: str = "", port: int = 0, timeout: float = 3.0, protocol: Literal["tcp", "udp"] = "tcp") -> bool:
    """

    Wait for a port to open. Useful when writing scripts which need to wait for a server to be available.

    :param host: hostname or ipaddress
    :param port: port
    :param timeout: timeout seconds (float)
    :param protocol: tcp or udp
    :return:

    Example:

        .. code-block:: python

            from pawnlib.resource import net
            net.wait_for_port_open("127.0.0.1", port)

            ## ⠏  Wait for port open 127.0.0.1:9900 ... 6


    """
    message = f"[bold green] Wait for port open {host}:{port} ..."
    count = 0
    with pawn.console.status(message) as status:
        while True:
            if check_port(host, port, timeout, protocol):
                status.stop()
                pawn.console.debug(f"[OK] Activate port -> {host}:{port}")
                pawn.app_logger.info(f"[OK] Activate port -> {host}:{port}")
                return True
            status.update(f"{message} {count}")
            count += 1
            time.sleep(1)



import os
import sys
import time
import inspect
import errno
import signal
import atexit
import subprocess
import threading
import itertools
from io import TextIOWrapper
from typing import Callable, List, Dict, Union
from concurrent.futures import ThreadPoolExecutor

from pawnlib.output import dump, debug_print, bcolors
from pawnlib import typing
from functools import wraps

from pawnlib.config.globalconfig import pawnlib_config as pawn
from pawnlib.typing.converter import shorten_text


class ThreadPoolRunner:
    """
    A class that runs a function with multiple arguments in parallel using a thread pool.

    :param func: The function to run in parallel.
    :type func: function
    :param tasks: A list of arguments to pass to the function.
    :type tasks: list
    :param max_workers: The maximum number of worker threads to use.
    :type max_workers: int
    :param verbose: Whether to print the results of each task as they complete.
    :type verbose: int
    :param sleep: The number of seconds to sleep between runs when using `forever_run`.
    :type sleep: int

    Example:

        .. code-block:: python

            runner = ThreadPoolRunner(func=my_function, tasks=my_args, max_workers=10, verbose=1, sleep=5)
            results = runner.run()
            runner.forever_run()

    """

    def __init__(self, func=None, tasks=[], max_workers=20, verbose=0, sleep=1):
        self.func = func
        self.tasks = tasks
        self.max_workers = max_workers
        self.results = []
        self.sleep = sleep
        self.verbose = verbose

    def initializer_worker(self):
        """
        A method that is run once by each worker thread when the thread pool is created.
        """
        pass

    def run(self):
        """
        Run the function with the given arguments in parallel using a thread pool.

        :return: A list of results from each task.
        :rtype: list
        """
        with ThreadPoolExecutor(max_workers=self.max_workers, initializer=self.initializer_worker) as pool:
            results = pool.map(self.func, self.tasks)
            if self.verbose > 0:
                self.log_results(results)
        return results

    @staticmethod
    def log_results(results):
        """
        Print the results of each task as they complete.

        :param results: A list of results from each task.
        :type results: list
        """
        for result in results:
            if result:
                print(result)

    def forever_run(self):
        """
        Run the function with the given arguments in parallel using a thread pool indefinitely.
        """
        while True:
            self.run()
            time.sleep(self.sleep)


class Daemon(object):
    """
    A generic daemon class.
    Usage 1: subclass the Daemon class and override the run() method
    Usage 2: subclass the Daemon class and use func parameter

    :param pidfile: pid file location
    :param func: function to run as daemon
    :param stdin: standard input , The default is sys.stdin, and providing a filename will output to a file.
    :param stdout: standard output, The default is sys.stdout, and  providing a filename will output to a file.
    :param stderr: standard error, The default is sys.stderr, and providing a filename will output to a file.
    :param home_dir: home directory
    :param umask: umask
    :param verbose: verbosity level
    :param use_gevent: use gevent
    :param use_eventlet: use eventlet
    Example:

        .. code-block:: python

            from pawnlib.utils.operate_handler import Daemon

            def main():
                while True:
                    print(f"main loop")
                    print("start daemon")
                    time.sleep(5)


            if __name__ == "__main__":
                if len(sys.argv) != 2:
                    sys.exit()
                command = sys.argv[1]
                daemon = Daemon(
                    pidfile="/tmp/jmon_agent.pid",
                    func=main
                )
                if command == "start":
                    daemon.start()
                elif command == "stop":
                    daemon.stop()
                else:
                    print("command not found [start/stop]")


    """

    def __init__(self, pidfile, func=None, stdin=None,
                 stdout=None, stderr=None,
                 home_dir='.', umask=0o22, verbose=1,
                 use_gevent: bool = False, use_eventlet: bool = False):
        self.stdin = stdin if stdin is not None else sys.stdin
        self.stdout = stdout if stdout is not None else sys.stdout
        self.stderr = stderr if stderr is not None else sys.stderr
        self.pidfile = pidfile
        self.func = func
        self.home_dir = home_dir
        self.verbose = verbose
        self.umask = umask
        self.daemon_alive = True
        self.use_gevent = use_gevent
        self.use_eventlet = use_eventlet

    def log(self, *args):
        if self.verbose >= 1:
            print(*args)

    def daemonize(self):
        """
        Do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """
        if self.use_eventlet:
            import eventlet.tpool
            eventlet.tpool.killall()
        try:
            pid = os.fork()
            if pid > 0:
                # Exit first parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write(
                "fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # Decouple from parent environment
        os.chdir(self.home_dir)
        os.setsid()
        os.umask(self.umask)

        # Do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # Exit from second parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write(
                "fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        if sys.platform != 'darwin':  # This block breaks on OS X
            # Redirect standard file descriptors
            sys.stdout.flush()
            sys.stderr.flush()
            si = open(self.stdin, 'r') if isinstance(self.stdin, str) else self.stdin
            so = open(self.stdout, 'a+') if isinstance(self.stdout, str) else self.stdout
            if self.stderr:
                try:
                    se = open(self.stderr, 'a+') if isinstance(self.stderr, str) else self.stderr
                except ValueError:
                    # Python 3 can't have unbuffered text I/O
                    se = open(self.stderr, 'a+', 1)
            else:
                se = so
            if isinstance(si, TextIOWrapper):
                os.dup2(si.fileno(), sys.stdin.fileno())
            if isinstance(so, TextIOWrapper):
                os.dup2(so.fileno(), sys.stdout.fileno())
            if isinstance(se, TextIOWrapper):
                os.dup2(se.fileno(), sys.stderr.fileno())

        def sigtermhandler(signum, frame):
            self.daemon_alive = False
            sys.exit()

        if self.use_gevent:
            import gevent
            gevent.reinit()
            gevent.signal(signal.SIGTERM, sigtermhandler, signal.SIGTERM, None)
            gevent.signal(signal.SIGINT, sigtermhandler, signal.SIGINT, None)
        else:
            signal.signal(signal.SIGTERM, sigtermhandler)
            signal.signal(signal.SIGINT, sigtermhandler)

        self.log("Started")

        # Write pidfile
        atexit.register(
            self.delpid)  # Make sure pid file is removed if we quit
        pid = str(os.getpid())
        open(self.pidfile, 'w+').write("%s\n" % pid)

    def delpid(self):
        try:
            # the process may fork itself again
            pid = int(open(self.pidfile, 'r').read().strip())
            if pid == os.getpid():
                os.remove(self.pidfile)
        except OSError as e:
            if e.errno == errno.ENOENT:
                pass
            else:
                raise

    def start(self, *args, **kwargs):
        """
        Start the daemon
        """

        self.log("Starting...")

        # Check for a pidfile to see if the daemon already runs
        try:
            pf = open(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None
        except SystemExit:
            pid = None

        if pid:
            message = "pidfile %s already exists. Is it already running?\n"
            sys.stderr.write(message % self.pidfile)
            sys.exit(1)

        # Start the daemon
        self.daemonize()
        if self.func:
            self.func(*args, **kwargs)
        else:
            self.run(*args, **kwargs)

    def stop(self):
        """
        Stop the daemon
        """

        if self.verbose >= 1:
            self.log("Stopping...")

        # Get the pid from the pidfile
        pid = self.get_pid()

        if not pid:
            message = "pidfile %s does not exist. Not running?\n"
            sys.stderr.write(message % self.pidfile)

            # Just to be sure. A ValueError might occur if the PID file is
            # empty but does actually exist
            if os.path.exists(self.pidfile):
                os.remove(self.pidfile)

            return  # Not an error in a restart

        # Try killing the daemon process
        try:
            i = 0
            while 1:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
                i = i + 1
                if i % 10 == 0:
                    os.kill(pid, signal.SIGHUP)
        except OSError as err:

            if os.path.exists(self.pidfile):
                os.remove(self.pidfile)
            else:
                print(str(err))
                sys.exit(0)

        self.log("Stopped")

    def restart(self):
        """
        Restart the daemon
        """
        self.stop()
        self.start()

    def get_pid(self):
        try:
            pf = open(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None
        except SystemExit:
            pid = None
        return pid

    def is_running(self):
        pid = self.get_pid()

        if pid is None:
            self.log('Process is stopped')
            return False
        elif os.path.exists('/proc/%d' % pid):
            self.log('Process (pid %d) is running...' % pid)
            return True
        else:
            self.log('Process (pid %d) is killed' % pid)
            return False

    def run(self):
        """
        You should override this method when you subclass Daemon.
        It will be called after the process has been
        daemonized by start() or restart().
        """
        raise NotImplementedError


def timing(f):
    """
    Get the time taken to complete the task.

    :param f:
    :return:
    """

    @wraps(f)
    def wrap(*args, **kwargs):
        function_name = f"{f.__module__}.{f.__name__}()"
        job_start(function_name)
        start_time = time.time()
        ret = f(*args, **kwargs)
        end_time = time.time() - start_time
        # cprint('{:s} function took {:.3f} ms'.format(f.__name__, (time2-time1)*1000.0))
        job_done(full_module_name=function_name, elapsed=end_time)
        return ret
    return wrap


def get_inspect_module(full_module_name=None):
    if full_module_name is None:
        module_name = ''
        stack = inspect.stack()
        parent_frame = stack[1][0]
        module = inspect.getmodule(parent_frame)
        if module:
            module_pieces = module.__name__.split('.')
            module_name = typing.list_to_oneline_string(module_pieces)
        function_name = stack[1][3]
        full_module_name = "%s.%s()" % (module_name, function_name)
    return full_module_name


def job_start(full_module_name=None):
    full_module_name = get_inspect_module(full_module_name)
    if pawn.get('PAWN_VERBOSE', 0) > 0:
        debug_print(f"[START] {full_module_name}", "green")


def job_done(error: str = '', full_module_name: str = '', elapsed: Union[float, int] = 0):
    title = get_inspect_module(full_module_name)
    if pawn.get('PAWN_VERBOSE', 0) > 0:
        if error == '':
            debug_print(f"[DONE ] {title} {elapsed:.3f}sec", "green")
        else:
            debug_print(" NOT DONE %s (%.3fsec)- %s" % (title, elapsed, error), "red")
    return title


def execute_function(module_func):
    if "." in module_func:
        [module_name, function_name] = module_func.split(".")
        dump(globals())
        module = __import__(f"{module_name}")
        func = getattr(module, function_name)
        return func()
    return globals()[module_func]()


def run_execute(text=None, cmd=None, cwd=None, check_output=True, capture_output=True, hook_function=None, debug=False, **kwargs):
    """
    Helps run commands

    :param text: just a title name
    :param cmd: command to be executed
    :param cwd: the function changes the working directory to cwd
    :param check_output:
    :param capture_output:
    :param hook_function:
    :param debug:
    :return:
    """

    if cmd is None:
        cmd = text

    start = time.time()

    result = dict(
        stdout=[],
        stderr=None,
        return_code=0,
        line_no=0
    )

    if text != cmd:
        text = f"text='{text}', cmd='{cmd}' :: "
    else:
        text = f"cmd='{cmd}'"

    try:
        process = subprocess.Popen(cmd, universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd, shell=True)

        for line in process.stdout:
            line_striped = line.strip()
            if line_striped:
                if callable(hook_function):
                    if hook_function == print:
                        print(f"[{result['line_no']}] {line_striped}")
                    else:
                        hook_function(line=line_striped, line_no=result['line_no'], **kwargs)

                if capture_output:
                    result["stdout"].append(line_striped)
                result['line_no'] += 1

        out, err = process.communicate()

        if process.returncode:
            result["return_code"] = process.returncode
            result["stderr"] = err.strip()

    except Exception as e:
        result['stderr'] = e
        raise OSError(f"Error while running command cmd='{cmd}', error='{e}'")

    end = round(time.time() - start, 3)

    if check_output:
        if result.get("stderr"):
            # cprint(f"[FAIL] {text}, Error = '{result.get('stderr')}'", "red")
            pawn.error_logger.info(f"[FAIL] {text}, Error = '{result.get('stderr')}'") if pawn.error_logger else False
        else:
            # cprint(f"[ OK ] {text}, timed={end}", "green")
            pawn.app_logger.info(f"[ OK ] {text}, timed={end}") if pawn.app_logger else False
    return result


def hook_print(*args, **kwargs):
    """
    Print to output every 10th line

    :param args:
    :param kwargs:
    :return:
    """
    if "amplify" in kwargs.get("line"):
        print(f"[output hook - matching keyword] {args} {kwargs}")

    if kwargs.get("line_no") % 100 == 0:
        print(f"[output hook - matching line_no] {args} {kwargs}")
    # print(kwargs.get('line'))


class Spinner:
    """
    Create a spinning cursor

    :param text: text
    :param delay: sleep time

    :Example

        .. code-block:: python

            from pawnlib.utils.operate_handler import Spinner
            with Spinner(text="Wait message"):
                time.sleep(10)

    """
    def __init__(self, text="", delay=0.1):

        self._spinner_items = itertools.cycle(['-', '/', '|', '\\'])
        self.delay = delay
        self.busy = False
        self.spinner_visible = False
        self.text = text
        self._screen_lock = None
        self.thread = None
        self.spin_message = ""
        self.line_up = '\033[1A'
        self.line_up = '\x1b[1A'
        self.line_clear = '\x1b[2K'

        if type(sys.stdout).__name__ == "FileProxy":
            self._sys_stdout = getattr(sys.stdout, "rich_proxied_file", sys.stdout)
        else:
            self._sys_stdout = sys.stdout

    def title(self, text=None):
        # print(end=self.line_up)
        self.text = text

    def write_next(self):
        with self._screen_lock:
            if not self.spinner_visible:
                if self.text:
                    self.spin_message = f"{self.text} ... {next(self._spinner_items)}"
                else:
                    self.spin_message = next(self._spinner_items)
                # sys.stdout.write(next(self.spinner))
                self._sys_stdout.write(self.spin_message)

                self.spinner_visible = True
                self._sys_stdout.flush()

    def remove_spinner(self, cleanup=False):
        with self._screen_lock:
            if self.spinner_visible:
                b = len(self.spin_message)
                self._sys_stdout.write('\b' * b)
                self.spinner_visible = False
                if cleanup:
                    self._sys_stdout.write(' ')       # overwrite spinner with blank
                    self._sys_stdout.write('\r')      # move to next line
                self._sys_stdout.flush()

    def spinner_task(self):
        while self.busy:
            self.write_next()
            time.sleep(self.delay)
            self.remove_spinner()

    def start(self):
        self._screen_lock = threading.Lock()
        self.busy = True
        self.thread = threading.Thread(target=self.spinner_task)
        self.thread.start()

    def stop(self):
        self.busy = False
        self.remove_spinner(cleanup=True)

    def __enter__(self):
        if self._sys_stdout.isatty():
            self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_traceback):
        if self._sys_stdout.isatty():
            self.stop()
            print(f"[DONE] {self.text}")
        else:
            self._sys_stdout.write('\r')


class WaitStateLoop:
    """
    loop_function is continuously executed and values ​​are compared with exit_function.

    :param loop_function:  function to run continuously
    :param exit_function: function to exit the loop
    :param timeout:
    :param delay: sleep time
    :param text: text message


    Example:

        .. code-block:: python

            from pawnlib.utils.operate_handler import WaitStateLoop
            from functools import partial

            def check_func(param=None):
                time.sleep(0.2)
                random_int = random.randint(1, 1000)
                # print(f"param= {param}, random_int = {random_int}")
                return random_int

            def loop_exit_func(result):
                if result % 10 == 1.5:
                    return True
                return False

            WaitStateLoop(
                loop_function=partial(check_func, "param_one"),
                exit_function=loop_exit_func,
                timeout=10
            ).run()

    """
    def __init__(self,
                 loop_function: Callable,
                 exit_function: Callable,
                 timeout=30, delay=0.5, text="WaitStateLoop",
                 ):
        self.loop_function = loop_function
        self.exit_function = exit_function
        self.timeout = timeout
        self.delay = delay
        self.text = text

    def run(self):
        """
        run()

        :return:
        """
        spin_text = ""
        error_text = ""
        count = 0
        start_time = time.time()
        if getattr(self.loop_function, "func"):
            func_name = self.loop_function.func.__name__
            spin_text = f"[{self.text}] Wait for {func_name}{self.loop_function.args}"

        with Spinner(text=spin_text) as spinner:
            while True:
                result = self.loop_function()
                is_success = self.exit_function(result)
                elapsed = int(time.time() - start_time)
                # spinner.title(f"{error_text}[{count}]{spin_text}: result={result}, is_success={is_success}, {elapsed} {time.time()} < {start_time + self.timeout}")
                if is_success is True:
                    return result
                spinner.title(f"{error_text}[{count}]{spin_text}: result={shorten_text(result, 30)}, is_success={is_success}, {elapsed} secs passed")
                # spinner.title(f" {elapsed} secs passed")
                time.sleep(self.delay)

                try:
                    assert time.time() < start_time + self.timeout
                except AssertionError:
                    # text = f"[{count:.1f}s] [{timeout_limit}s Timeout] Waiting for {exec_function_name} / '{func_args}' :: '{wait_state}' -> {check_state} , {error_msg}"
                    error_text = f"[TIMEOUT]"
                count += 1


def wait_state_loop(
        exec_function=None,
        func_args=[],
        check_key="status",
        wait_state="0x1",
        timeout_limit=30,
        increase_sec=0.5,
        health_status=None,
        description="",
        force_dict=True,
        logger=None
):
    start_time = time.time()
    count = 0
    # arguments 가 한개만 있을 때의 예외
    if isinstance(func_args, str):
        tmp_args = ()
        tmp_args = tmp_args + (func_args,)
        func_args = tmp_args

    exec_function_name = exec_function.__name__
    act_desc = f"desc={description}, function={exec_function_name}, args={func_args}"
    spinner = Halo(text=f"[START] Wait for {description} , {exec_function_name}, {func_args}", spinner='dots')
    if logger and hasattr(logger, "info"):
        logger.info(f"[SR] [START] {act_desc}")

    spinner.start()

    while True:
        if isinstance(func_args, dict):
            response = exec_function(**func_args)
        else:
            response = exec_function(*func_args)

        if not isinstance(response, dict):
            response = response.__dict__

        if force_dict and isinstance(response.get("json"), list):
            response['json'] = response['json'][0]

        check_state = ""
        error_msg = ""

        if response.get("json") or health_status:
            response_result = response.get("json")
            check_state = response_result.get(check_key, "")
            response_status = response.get("status_code")
            if check_state == wait_state or health_status == response_status:
                status_header = bcolors.OKGREEN + "[DONE]" + bcolors.ENDC
                text = f"\t[{description}] count={count}, func={exec_function_name}, args={str(func_args)[:30]}, wait_state='{wait_state}', check_state='{check_state}'"
                if health_status:
                    text += f", health_status={health_status}, status={response_status}"
                spinner.succeed(f'{status_header} {text}')
                spinner.stop()
                spinner.clear()
                # spinner.stop_and_persist(symbol='🦄'.encode('utf-8'), text="[DONE]")
                break
            else:
                if type(response_result) == dict or type(check_state) == dict:
                    if response_result.get("failure"):
                        if response_result.get("failure").get("message"):
                            print("\n\n\n")
                            spinner.fail(f'[FAIL] {response_result.get("failure").get("message")}')
                            spinner.stop()
                            spinner.clear()
                            break

        text = f"[{count:.1f}s] Waiting for {exec_function_name} / {func_args} :: '{wait_state}' -> '{check_state}' , {error_msg}"
        spinner.start(text=text)

        if logger and hasattr(logger, "info"):
            logger.info(f"[SR] {text}")

        try:
            assert time.time() < start_time + timeout_limit
        except AssertionError:
            text = f"[{count:.1f}s] [{timeout_limit}s Timeout] Waiting for {exec_function_name} / '{func_args}' :: '{wait_state}' -> {check_state} , {error_msg}"
            spinner.start(text=text)

            if logger and hasattr(logger, "error"):
                logger.info(f"[SR] {text}")

        count = count + increase_sec
        time.sleep(increase_sec)

        spinner.stop()

    if logger and hasattr(logger, "info"):
        logger.info(f"[SR] [DONE] {act_desc}")

    if health_status:
        return response


def run_with_keyboard_interrupt(command, *args, **kwargs):
    """
    run with KeyboardInterrupt
    :param command:
    :param args:
    :param kwargs:
    :return:


    Example:

        .. code-block:: python

            from pawnlib.utils.operate_handler import run_with_keyboard_interrupt

            run_with_keyboard_interrupt(run_func, args, kwargs)


    """
    try:
        if callable(command):
            command(*args, **kwargs)
        else:
            pawn.console.print(f"\n[red] {command} not callable ")
    except KeyboardInterrupt:
        pawn.console.print(f"\n\n[red] ^C KeyboardInterrupt - {command.__name__}{str(args)[:-1]}{kwargs}) \n")


def handle_keyboard_interrupt_signal():
    import signal

    def handle_ctrl_c(_signal, _frame):
        pawn.console.rule(f"[red] KeyboardInterrupt, Going down! Signal={_signal}")
        sys.exit(0)
    signal.signal(signal.SIGINT, handle_ctrl_c)



#!/usr/bin/env python3
import argparse
from pawnlib.builder.generator import generate_banner
from pawnlib.__version__ import __version__ as _version
from pawnlib.config import pawnlib_config as pawn, pconf
from pawnlib.utils import disable_ssl_warnings

from pawnlib.resource import net

disable_ssl_warnings()
NONE_STRING = "__NOT_DEFINED_VALUE__"
PLATFORM_LIST = ["icon", "havah"]


def get_parser():
    parser = argparse.ArgumentParser(description='RPC')
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument(
        'command',
        help='check',
        nargs='?'
    )
    parser.add_argument('-d', '--debug', action='store_true', help='debug mode. True/False')
    parser.add_argument('-v', '--verbose', action='count', help='verbose mode. view level', default=0)
    return parser


def print_banner():
    banner = generate_banner(
        app_name="NET",
        author=f"jinwoo",
        description="Network checker",
        font="graffiti",
        version=_version
    )
    print(banner)


def initialize_arguments():
    parser = get_parser()
    args, unknown = parser.parse_known_args()
    args.subparser_name = "net"
    pawn.set(
        PAWN_DEBUG=args.debug,
        data=dict(
            args=args
        )
    )
    print_banner()
    return args


def main():
    args = initialize_arguments()
    disable_ssl_warnings()

    ff_region = net.FindFastestRegion(verbose=args.verbose)
    ff_region.run()
    ff_region.print_results()


if __name__ == '__main__':
    main()



from .net import (
    OverrideDNS,
    get_public_ip,
    get_local_ip,
    get_hostname,
    check_port,
    listen_socket,
    wait_for_port_open
)

from .server import (
    SystemMonitor,
    get_interface_ips_dict,
    get_interface_ips,
    get_ip_and_netmask,
    subnet_mask_to_decimal,
    get_default_route_and_interface,
    get_cpu_usage_percentage,
    get_rlimit_nofile,
    get_platform_info,
    get_cpu_load,
    get_mem_info,
    get_uptime_cmd,
    get_total_memory_usage,
    get_mac_platform_info,
    get_mem_osx_info,
    get_cpu_time,
    get_aws_metadata,
    aws_data_crawl,
    get_netstat_count,

)



#!/usr/bin/env python3
import argparse
from pawnlib.builder.generator import generate_banner
from pawnlib.__version__ import __version__ as _version
from pawnlib.output.color_print import *
from pawnlib.output import write_yaml, is_file, open_yaml_file
from pawnlib.config import pawnlib_config as pawn, pconf
from pawnlib.utils import IconRpcHelper, IconRpcTemplates, NetworkInfo, icx_signer

from InquirerPy import prompt
from pawnlib.utils import disable_ssl_warnings
from pawnlib.input import PromptWithArgument, PrivateKeyValidator, StringCompareValidator, PrivateKeyOrJsonValidator

disable_ssl_warnings()
NONE_STRING = "__NOT_DEFINED_VALUE__"
PLATFORM_LIST = ["icon", "havah"]


def get_parser():
    parser = argparse.ArgumentParser(description='RPC')
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument(
        'command',
        help='account, icx_sendTransaction, icx_sendTransaction_v3, get_transactionResult, icx_getBalance, icx_getTotalSupply',
        nargs='?'
    )
    parser.add_argument('--url', metavar='url', help='endpoint url default: None', default=None)
    parser.add_argument('--from', metavar='address', dest='from', help='from address. default: None', default=None)
    parser.add_argument('--to', metavar='address', dest="to", help='to address. default: None', default=None)
    parser.add_argument('--address', metavar='address', help=f'icx address. default: None', default=None)
    parser.add_argument('--txhash', metavar='txhash', help='txhash')
    parser.add_argument('--value', metavar='amount', type=float, help=f'icx amount to transfer. unit: icx. ex) 1.0. default:0.001',
                        default=0.001)
    parser.add_argument('--fee', metavar='amount', type=float,
                        help='transfer fee. default: 0.01', default=0.001)
    parser.add_argument('--pk', metavar='private_key',
                        help=f'hexa string. default: None', default=None)
    parser.add_argument('--debug', action='store_true', help=f'debug mode. True/False')
    parser.add_argument('-n', '--number', metavar='number', type=int, help=f'try number. default: None', default=None)

    parser.add_argument('--nid', metavar='nid', type=str, help=f'network id default: None', default=None)

    parser.add_argument('-c', '--config', metavar='config', help='config name')

    parser.add_argument('-k', '--keystore', metavar='keystore', help='keystore file name')
    parser.add_argument('-p', '--password', metavar='password', help='keystore file password')

    parser.add_argument('-t', '--timeout', metavar='timeout', type=float, help=f'timeout')
    parser.add_argument('-w', '--worker', metavar='worker', type=int, help=f'worker')

    # parser.add_argument('-i', '--increase', metavar='increase_count', type=int, help=f'increase count number')
    # parser.add_argument('--increase-count', metavar='increase_count', type=int, help=f'increase count number', default=1)
    # parser.add_argument('-r', '--rnd_icx', metavar='rnd_icx', help=f'rnd_icx', default="no")

    parser.add_argument('-m', '--method', metavar='method', help='method for JSON-RPC', default="")
    parser.add_argument('--platform', type=lambda s : s.lower(), metavar='platform', help='platform name of network name',
                        # choices=PLATFORM_LIST,
                        # default="havah"
                        )

    parser.add_argument('--network', metavar='network_name', help='network name', default="")
    parser.add_argument('--fill-each-prompt',  action='store_true', help='fill each prompt', default=False)
    parser.add_argument('--base-dir', metavar='base_dir', help='base directory', default=os.getcwd())
    # parser.add_argument('--load-type', metavar='network_name', help='network name', default="")

    return parser


def get_methods(answers):
    icon_tpl = IconRpcTemplates()
    return icon_tpl.get_methods(answers['category'])


def get_required(answers):
    icon_tpl = IconRpcTemplates(category=answers['category'], method=answers['method'])
    pawn.console.log(f"get_required => {icon_tpl.get_required_params()}, {answers['category']}, {answers['method']}")

    return icon_tpl.get_required_params()


def fetch_environments_to_args():
    args = pconf().data.args
    env_map = {
        "PLATFORM": "platform",
        "SERVICE": "network",
    }
    for env_key, args_key in env_map.items():
        env_value = os.getenv(env_key, NONE_STRING)
        if env_value != NONE_STRING and \
                getattr(args, args_key, NONE_STRING) != NONE_STRING:
            pawn.console.log(f"[yellow]Defined from environment variable, {env_key}={env_value}, (args.{args_key})")
            setattr(args, args_key, env_value)


def print_banner():
    banner = generate_banner(
        app_name="RPC",
        author=f"jinwoo \n\n - Platform    : {str(pconf().data.args.platform).upper()}",
        description="JSON-RPC request",
        font="graffiti",
        version=_version
    )
    print(banner)


# def initialize_arguments():
#     parser = get_parser()
#     args, unknown = parser.parse_known_args()
#     args.subparser_name = "rpc"
#     pawn.set(
#         PAWN_DEBUG=args.debug,
#         data=dict(
#             args=args
#         )
#     )
#     print_banner()
#     fetch_environments_to_args()
#     pawn.console.log(args)
#
#
#     if args.platform not in PLATFORM_LIST:
#         raise ValueError(f'not supported platform, allowed = {PLATFORM_LIST}')
#
#     return args


def fill_sign_params_from_args(payload: dict):
    args = pconf().data.args
    args_list = ["to", "stepLimit", "nid", "value", "nonce"]

    payload.setdefault('params', {})

    for args_key in args_list:
        value = getattr(args, args_key, None)
        if value is not None:
            if args_key == "value":
                value = hex(int(value * 10 ** 18))
            pawn.console.debug(f"[yellow]Defined from parameters variable, {args_key}={value}, (args.{args_key})")
            payload['params'][args_key] = value
    return payload


def network_info_to_args(network_info):
    print(network_info)


class RpcCommand:
    def __init__(self):
        self.args = None
        self.network_info = None
        self.icon_tpl = IconRpcTemplates()
        self._payload = {}
        self.icon_rpc = None

        self.config_file = "config.yaml"
        self.config_data = {}

    def initialize_arguments(self):
        parser = get_parser()
        args, unknown = parser.parse_known_args()
        args.subparser_name = "rpc"
        pawn.set(
            PAWN_DEBUG=args.debug,
            data=dict(
                args=args
            )
        )
        print_banner()
        fetch_environments_to_args()
        pawn.console.log(args)
        self.load_config_file()
        self.args = pconf().data.args

        self.network_info = NetworkInfo(force=True)
        self.network_info.update_platform_info(self.config_data)

    def set_network_info(self):
        PromptWithArgument(
            message="Select Platform ?",
            choices=self.network_info.get_platform_list(),
            type="list",
            argument="platform",
        ).prompt()

        PromptWithArgument(
            message="Select Network ?",
            choices=self.network_info.get_network_list(platform=self.args.platform),
            type="list",
            argument="network",
            default="vega"
        ).prompt()

        self.network_info.set_network(platform=self.args.platform, network_name=self.args.network)
        pawn.console.log(self.network_info)

    def generate_tx_payload(self):
        self.icon_rpc = IconRpcHelper(network_info=self.network_info)
        pawn.console.log("Fetching Governance SCORE API")
        _governance_score_api  = self.icon_rpc.get_governance_api()
        self.icon_tpl.update_template(_governance_score_api)

        category = None
        if not self.args.method:
            category = PromptWithArgument(
                message="Select a category to use in JSON-RPC.",
                choices=self.icon_tpl.get_category(),
                long_instruction="\nUse the up/down keys to select",
                max_height="40%",
                default="",
                argument="",
                # args=pconf().data.args
            ).select()

        PromptWithArgument(
            message=">> Select a method to use in JSON-RPC.",
            choices=self.icon_tpl.get_methods(category=category),
            long_instruction="\nUse the up/down keys to select",
            type="list",
            max_height="40%",
            default="",
            argument="method",
            # args=pconf().data.args
        ).fuzzy()

        self._payload = self.icon_tpl.get_rpc(category=category, method=self.args.method)
        required_params = self.icon_tpl.get_required_params()

        if required_params and self.args.fill_each_prompt:
            _questions = []
            for k, v in required_params.items():
                _questions.append({'type': 'input', 'name': k.lower(), 'message': f'What\'s "{k}" parameter?'})
                # from 주소면 wallet 디렉토리를 읽어서 리스트를 보여준다.
            self._payload['params'] = prompt(_questions)

        if self.icon_tpl.is_required_sign() or self._payload.get('method') == "icx_sendTransaction":
            self.icon_rpc.wallet = icx_signer.WalletCli().load()
            self._payload['params']['from'] = self.icon_rpc.wallet.get('address')
            self.print_balance()
            self.prepare_signature()

        self._payload = PromptWithArgument(
            type="input",
            message="Edit transaction: ",
            default="\n"+json.dumps(self._payload, indent=4),
            long_instruction="\nedit the transaction",
        ).prompt()

        if self.icon_tpl.is_required_sign():
            self._payload = self.icon_rpc.sign_tx(payload=self._payload)
        self.icon_rpc.rpc_call(url=self.args.url, payload=self._payload)
        self.icon_rpc.print_request()
        self.icon_rpc.print_response(hex_to_int=True)

        if self.icon_tpl.is_required_sign():
            self.icon_rpc.get_tx_wait()

    def print_balance(self):
        address = self.icon_rpc.wallet.get('address')
        balance = self.icon_rpc.get_balance(address=address)
        pawn.console.log(f"address={address}, balance={balance} {self.network_info.symbol}")

    def prepare_signature(self):
        self._payload = fill_sign_params_from_args(self._payload)

        if self.network_info.nid:
            self._payload['params']['nid'] = self.network_info.nid

    def write_config_file(self):

        from pawnlib.output.file import check_file_overwrite
        if check_file_overwrite(filename=self.config_file):
            pawn.console.log("Write configuration config.yaml")
            network_info = NetworkInfo(force=True).get_platform_info()
            write_yaml(self.config_file, data=network_info)

    def load_config_file(self):
        if is_file(self.config_file):
            self.config_data = open_yaml_file(self.config_file)
            pawn.console.log(f"Loaded configuration from '{self.config_file}' len={len(self.config_data)}")
            pawn.console.log(self.config_data)

    def run(self):
        self.initialize_arguments()
        if self.args.command == "config":
            self.write_config_file()
            return

        self.set_network_info()
        self.generate_tx_payload()


def main():
    cli = RpcCommand()
    cli.run()


if __name__ == '__main__':
    print("*"* 100)
    main()



#!/usr/bin/env python3
import argparse
from pawnlib.builder.generator import generate_banner
from pawnlib.__version__ import __version__ as _version
from pawnlib.config import pawnlib_config as pawn
from pawnlib.output import write_json, syntax_highlight, PrintRichTable
from pawnlib.typing.converter import FlatDict, FlatterDict, flatten_dict

from pawnlib.resource import server

__description__ = 'Get meta information from AWS EC2.'


def get_parser():
    parser = argparse.ArgumentParser(description='AWS')
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument("--meta-ip", "-i", type=str, help="IP address for retrieving Metadata", default="169.254.169.254")
    parser.add_argument("--timeout", '-t', type=float, help="timeout for request", default=2)
    parser.add_argument("--write-filename", '-w', type=str, help="write filename", default="")
    parser.add_argument("--print-type", '-p', type=str, help="print type", choices=["json", "flat"], default="json")


    return parser


def main():
    banner = generate_banner(
        app_name="aws metadata",
        author="jinwoo",
        description="get the aws metadata",
        font="graffiti",
        version=_version
    )
    print(banner)

    parser = get_parser()
    args, unknown = parser.parse_known_args()

    pawn.console.log(f"args = {args}")

    res = server.get_aws_metadata(meta_ip=args.meta_ip, timeout=args.timeout)
    if args.print_type == "json":
        print(syntax_highlight(res))
    elif args.print_type == "flat":
        PrintRichTable(
            title="AWS Metadata",
            data=flatten_dict(res),
            columns_options=dict(
                value=dict(
                    justify="left",
                )
            )
        )

    # PrintRichTable(title="AWS Metadata", data=flatten_dict(res))

    if args.write_filename:
        write_res = write_json(filename=args.write_filename, data=res)
        pawn.console.log(write_res)


if __name__ == '__main__':
    main()



import os
import hashlib
import codecs


class KeyWallet:
    def __init__(self, private_key):
        self.private_key = private_key
        self.public_key = self._get_public_key()
        self.address = self.get_address()

    @staticmethod
    def create():
        # 새로운 임의의 개인 키 생성
        private_key = os.urandom(32)
        return KeyWallet(private_key)

    def _get_public_key(self):
        # 개인 키로부터 공개 키 유도
        public_key = self._point_multiplication(self.private_key, self._G)
        return public_key

    def get_public_key(self):
        # 개인 키로부터 공개 키 유도
        # return codecs.encode(self.public_key_to_bytes(), 'hex').decode()
        return codecs.encode(self.public_key_to_bytes(), 'hex').decode()

    def get_private_key(self):
        return codecs.encode(self.private_key, 'hex').decode()

    def get_address(self):
        # 공개 키로부터 지갑 주소 생성
        public_key_bytes = self.public_key_to_bytes()
        public_key_hash = hashlib.sha3_256(public_key_bytes).digest()
        wallet_address = 'hx' + codecs.encode(public_key_hash[-20:], 'hex').decode()
        return wallet_address

    def public_key_to_bytes(self):
        # 공개 키 튜플을 바이트로 변환
        x, y = self.public_key
        public_key_bytes = (x.to_bytes(32, 'big') + y.to_bytes(32, 'big'))
        return public_key_bytes

    def compressed_public_key(self):
        # Get the compressed public key (66 digits)
        x, y = self.public_key
        prefix = '03' if (y % 2) == 1 else '02'
        compressed_public_key = prefix + format(x, '064x')
        return compressed_public_key

    def _point_multiplication(self, scalar, point):
        # 타원 곡선에서의 점 곱셈 연산 수행
        result = self._point_at_infinity()
        for bit in scalar:
            if result is not None:
                result = self._point_double(result)
            if bit:
                if result is None:
                    result = point
                else:
                    result = self._point_add(result, point)
        return result

    def _point_double(self, point):
        # 타원 곡선에서의 점 두 배 연산 수행
        if point is None:
            return None
        x, y = point
        if y == 0:
            return self._point_at_infinity()
        slope = (3 * x * x) * self._inverse_mod(2 * y, self._p)
        x3 = (slope * slope - 2 * x) % self._p
        y3 = (slope * (x - x3) - y) % self._p
        return x3, y3

    def _point_add(self, point1, point2):
        # 타원 곡선에서의 점 덧셈 연산 수행
        if point1 is None:
            return point2
        if point2 is None:
            return point1
        x1, y1 = point1
        x2, y2 = point2
        if x1 == x2 and y1 != y2:
            return self._point_at_infinity()
        if x1 == x2:
            slope = (3 * x1 * x1) * self._inverse_mod(2 * y1, self._p)
        else:
            slope = (y1 - y2) * self._inverse_mod(x1 - x2, self._p)
        x3 = (slope * slope - x1 - x2) % self._p
        y3 = (slope * (x1 - x3) - y1) % self._p
        return x3, y3

    def _inverse_mod(self, a, m):
        # 모듈러 역수 계산
        if a < 0 or m <= a:
            a = a % m
        c, d = a, m
        uc, vc, ud, vd = 1, 0, 0, 1
        while c != 0:
            q, c, d = divmod(d, c) + (c,)
            uc, vc, ud, vd = ud - q * uc, vd - q * vc, uc, vc
        return ud % m

    def _point_at_infinity(self):
        # 타원 곡선에서의 무한 원점 반환
        return None

    # 타원 곡선 매개변수 정의 (secp256k1)
    _p = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F
    _G = (
        0x79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798,
        0x483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8
    )

# 테스트 수행




from functools import wraps, partial
import asyncio
import aiometer
from aiometer._impl import utils
from pawnlib.output import debug_print, classdump
from pawnlib.config import pawnlib_config as pawn
from pawnlib.utils.http import ALLOWS_HTTP_METHOD
import httpx


class AsyncTasks:
    def __init__(self,
                 max_at_once: int = 10,
                 max_per_second: int = 10,
                 title: str = "Working on async tasks ...",
                 debug: bool = False,
                 status: bool = False,
                 **kwargs):
        """
        This Class is to run asyncio using aiometer.

        :param max_at_once: Limit maximum number of concurrently running tasks.
        :param max_per_second: Limit request rate to not overload the server.
        :param title: Title of the tasks
        :param status: Status of the tasks
        :param debug: Whether to use debug
        :param kwargs:

        Example:

            .. code-block:: python

                async_tasks = AsyncTasks(max_at_once=10, max_per_second=10)
                async_tasks.generate_tasks(
                    target_list=range(1, 100) ,
                    function=run_container,
                    **{"args": args}
                ).run()


        """
        self.tasks = []
        self.max_at_once = max_at_once
        self.max_per_second = max_per_second
        self.debug = debug
        self._debug_print(self)

        self.get_list_function = None

        self.async_partial_target_func = None
        self.async_partial_task_func = None
        self.status_console = None

        self._title = title
        self._view_status = status
        self._function_name = ""

    def get_tasks(self):
        return self.tasks

    def generate_tasks(self, target_list=None, function=None, **kwargs):
        """
        This function generates the async tasks list

        :param target_list: List of targets for asynchronous execution
        :param function: Name of the function to execute
        :param kwargs:
        :return:
        """
        if function and getattr(function, "__qualname__"):
            self._function_name = function.__qualname__
        else:
            raise ValueError(f"{function} is not function")

        self._debug_print(f"{target_list}, {type(target_list)}")
        if kwargs.get('kwargs'):
            kwargs = kwargs['kwargs']
        else:
            kwargs = {}

        if target_list is None:
            target_list = []

        for target in target_list:
            self._debug_print(f"target={target}, function={self._function_name}(), kwargs={kwargs}, task_len={len(self.tasks)}")
            self.tasks.append(partial(function, target, **kwargs))

    def run(self):
        """
        This function executes an asynchronous operation.

        :return:
        """

        if self._view_status:
            with pawn.console.status(self._title) as status:
                result = asyncio.run(self._runner(status))
        else:
            result = asyncio.run(self._runner())

        return result

    async def _runner(self, status=None):
        result = {}
        tasks_length = len(self.tasks)

        if tasks_length > 0:
            async with aiometer.amap(
                    async_fn=lambda fn: fn(),
                    args=self.tasks,
                    max_at_once=self.max_at_once,
                    max_per_second=self.max_per_second,
                    _include_index=True,
            ) as amap_results:
                async for _index, _result in amap_results:
                    result[_index] = _result
                    if status and self._view_status:
                        status.update(f"{self._title} <{self._function_name}> [{_index}/{tasks_length}] {_result}")
        else:
            pawn.console.log(f"ERROR: tasks is null = {self.tasks}")
        return utils.list_from_indexed_dict(result)

    def _debug_print(self, *args, **kwargs):
        if self.debug:
            debug_print(*args, **kwargs)


def async_partial(async_fn, *args, **kwargs):
    async def wrapped():
        result = None
        if asyncio.iscoroutinefunction(async_fn):
            result = await async_fn(*args, **kwargs)
        else:
            debug_print(f"{async_fn} is not coroutine", "red")
        return result
    return wrapped


def run_in_async_loop(f):
    @wraps(f)
    async def wrapped(*args, **kwargs):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, f(*args, **kwargs))
    return wrapped


class AsyncHttp(AsyncTasks):
    """
    This class is a subclass of AsyncTasks and is used to handle asynchronous HTTP requests.

    Attributes:
        max_at_once (int): Maximum number of tasks to run at once.
        max_per_second (int): Maximum number of tasks to start per second.
        title (str): Title for the progress bar.
        debug (bool): If True, print debug information.
        status (bool): If True, print status information.
        urls (list): List of URLs to fetch.
    """

    def __init__(self,
                 max_at_once: int = 10,
                 max_per_second: int = 10,
                 title="Working on async tasks ...",
                 debug: bool = False,
                 status: bool = False,
                 urls: list = None,
                 **kwargs):
        if urls is None:
            urls = []
        super().__init__(max_at_once, max_per_second, title, debug, status, **kwargs)
        self.urls = urls
        self._prepare()

    def append_task(self, task):
        """
        Append a new task to the task list.

        Args:
            task (str or dict): If it's a string, it's considered as a URL. If it's a dictionary, it should contain a 'url' key.
        """

        _url = None
        _kwargs = {}

        if isinstance(task, str):
            _url = task
            _kwargs = {}
        elif isinstance(task, dict) and task.get('url'):
            _url = task.pop("url")
            _kwargs = task

        self.generate_tasks(
            target_list=[_url],
            function=fetch_httpx_url,
            **_kwargs
        )
        pawn.console.log(f"IN] url={_url}, kwargs={_kwargs}, max_at_once={self.max_at_once}")

    def _prepare(self):
        """
        Prepare the tasks based on the provided URLs.
        """

        if len(self.urls) > 0:
            for info in self.urls:

                self.append_task(info)


async def fetch_httpx_url(url, method="get", timeout=4, info="", max_keepalive_connections=10, max_connections=20, **kwargs):
    """
    Asynchronously fetch a URL using httpx.

    Args:
        url (str): The URL to fetch.
        method (str): The HTTP method to use. Default is "get".
        timeout (int): The timeout for the request in seconds. Default is 4.
        info (str): Additional information for the request. Default is an empty string.
        max_keepalive_connections (int): The maximum number of keep-alive connections. Default is 10.
        max_connections (int): The maximum number of connections. Default is 20.
        **kwargs: Additional keyword arguments for the httpx request.

    Returns:
        httpx.Response or dict: The response from the server, or an empty dictionary if an error occurred.
    """

    response = None
    try:
        limits = httpx.Limits(max_keepalive_connections=max_keepalive_connections, max_connections=max_connections)
        async with httpx.AsyncClient(timeout=timeout, limits=limits) as client:

            # pawn.console.log(url, kwargs)

            if method in ALLOWS_HTTP_METHOD:
                response = await getattr(client, method)(url, timeout=timeout, **kwargs)
            else:
                pawn.console.log(f"[ERROR] Unsupported HTTP method -> {method}")
                pawn.app_logger.error(f"[ERROR] Unsupported HTTP method -> {method}")
                raise ValueError(f"[ERROR] Unsupported HTTP method -> {method}")

            # client.post

            #
            # if method == "get":
            #     response = await client.get(url)
            # else:
            #     response = await getattr(client, method)(url)

            if response.status_code != 200:
                pawn.console.log(f"[red][ERROR] fetching {url}, status_code={response.status_code}, response={response.text}")
                pawn.app_logger.error(f"[red][ERROR] fetching {url}, status_code={response.status_code}, response={response.text}")
            return response

    except Exception as e:
        pawn.console.log(f"url={url} e={e}, response={response}, info={info}")
        pawn.app_logger.error(f"url={url} e={e}, response={response}, info={info}")
    return {}



from pawnlib.config import pawnlib_config as pawn
from pawnlib.output import get_real_path, classdump
import argparse

from glob import glob
import os
import sys
import importlib
from pawnlib.builder.generator import generate_banner
from pawnlib.__version__ import __version__ as _version
from pawnlib.utils.operate_handler import run_with_keyboard_interrupt
from pawnlib.typing.check import sys_exit


def load_submodule_parsers(parent_module, parser, help=None):
    if help is None:
        help = parent_module.__name__ + " modules"
    modules = glob(os.path.join(parent_module.__path__, "*.py"))
    subparsers = parser.add_subparsers(help=help)
    for module_file in modules:
        module_name = os.path.basename(module_file)[:-3]
        if module == "__init__":
            continue
        module = importlib.import_module(module_name, package=parent_module.__name__)
        if "define_arguments" not in module.__dict__:
            raise ImportError(parent_module.__name__ + " submodule '" + module_name + "' does not have required 'define_arguments' function.")
        parser = subparsers.add_parser(module_name)
        module.define_arguments(parser)


def load_all_submodule():
    for module_file in get_submodule_names():
        module_name = module_file
        pawn.console.log(module_name)
        run_module(module_name)


def get_module_name(name):
    return os.path.basename(name)[:-3]


def get_submodule_names():
    module_list = glob(os.path.join(get_real_path(__file__), "*.py"))
    main_cli_name = get_module_name(__file__)
    exclude_module_names = ["__init__", main_cli_name]
    modules = []
    for module_file in module_list:
        module_name = get_module_name(module_file)
        if module_name not in exclude_module_names:
            modules.append(module_name)
    return modules


def run_module(module_name=None):
    module = importlib.import_module(f"pawnlib.cli.{module_name}")
    module.main()
    pawn.console.debug(f"load a {module_name}")
    return module


def remove_argument(parser, arg):
    for action in parser._actions:
        opts = action.option_strings
        if (opts and opts[0] == arg) or action.dest == arg:
            parser._remove_action(action)
            pawn.console.log("ok remove")
            break

    for action in parser._action_groups:
        for group_action in action._group_actions:
            if group_action.dest == arg:
                action._group_actions.remove(group_action)
                pawn.console.log("ok remove")
                return


def parse_args(parser, commands):
    split_argv = [[]]
    if len(sys.argv) > 1:
        command = sys.argv[1]
    else:
        command = None
    for c in sys.argv[1:]:
        if c in commands.choices:
            split_argv.append([c])
        else:
            split_argv[-1].append(c)
    args = argparse.Namespace()
    for c in commands.choices:
        setattr(args, c, None)
    parser.parse_args(split_argv[0], namespace=args)  # Without command
    for argv in split_argv[1:]:  # Commands
        n = argparse.Namespace()
        setattr(args, argv[0], n)
        parser.parse_args(argv, namespace=n)
    return args, command


def get_sys_argv():
    if len(sys.argv) > 1:
        return sys.argv[1]
    return ""


def load_cli_module(commands=None, module_name=""):
    pawn.console.debug(f"Add parser => {module_name}")
    module = importlib.import_module(f"pawnlib.cli.{module_name}")
    description = getattr(module, "__description__", f"{module_name} module")
    _parser = commands.add_parser(module_name, help=f'{description}')
    module.get_arguments(_parser)


def get_args():
    parser = argparse.ArgumentParser(
        usage=generate_banner(app_name="PAWNS", version=_version, author="jinwoo", font="graffiti"),
        formatter_class=argparse.RawTextHelpFormatter
    )
    commands = parser.add_subparsers(title='sub-module')
    pawn.console.debug(f"sys_argv={get_sys_argv()}, modules={get_submodule_names()}")
    if get_sys_argv() and  get_sys_argv() in get_submodule_names():
        load_cli_module(commands, get_sys_argv())
    else:
        for module_name in get_submodule_names():
            try:
                load_cli_module(commands, module_name)
            except Exception as e:
                pawn.console.debug(f"[red] An error occurred while loading the module [/red] - {e}")

    args, command = parse_args(parser, commands)
    return args, command, parser


def cleanup_args():
    if len(sys.argv) > 2 and "-" not in sys.argv[2]:
        pawn.console.debug(f"Remove argument -> {sys.argv[1]}")
        del sys.argv[1]


def main():
    pawn.console.debug("Starting main_cli wrapper")
    args, command, parser = None, None, None
    try:
        pawn.console.debug(f"<before> {sys.argv}")
        args, command, parser = get_args()
        # pawn.console.debug(f"<after> parser {args}, {command}, {parser}")
        cleanup_args()
        # pawn.console.debug(f"<after> {sys.argv}")
    except Exception as e:
        pawn.console.debug(f"[red]Exception while parsing an argument = {e}")
        # sys_exit("Stopped CLI")
    pawn.console.debug(f"args={args}, command={command}, parser={parser}")

    if command:
        pawn.console.debug(f"command = {command}, PAWN_DEBUG={pawn.get('PAWN_DEBUG')}")
        try:
            # run_with_keyboard_interrupt(run_module, command)
            run_module(command)
        except KeyboardInterrupt:
            pawn.console.log("[red]KeyboardInterrupt")
        except Exception as e:
            if pawn.get('PAWN_DEBUG'):
                pawn.console.print_exception(show_locals=pawn.get("PAWN_DEBUG", False), width=160)
            else:
                pawn.console.log(f"[red]Exception -- {e}")
    else:
        if parser:
            parser.print_help()
        sys.exit(1)


if __name__ == '__main__':
    run_with_keyboard_interrupt(main)
    # main()




from .check import (
    is_json,
    is_float,
    is_int,
    is_hex,
    is_regex_keyword,
    is_regex_keywords,
    list_depth,
    is_valid_ipv4,
    is_valid_ipv6,
    is_valid_url,
    is_valid_private_key,
    is_valid_token_address,
    is_valid_tx_hash,
    guess_type,
    return_guess_type,
    sys_exit,
    is_include_list,
    detect_encoding,
    keys_exists,
)
from .defines import (
    Namespace,
    set_namespace_default_value,
    fill_required_data_arguments,
)
from .converter import (
    StackList,
    ErrorCounter,
    MedianFinder,
    FlatDict,
    MedianFinder,
    base64ify,
    base64_decode,
    UpdateType,
    convert_hex_to_int,
    convert_dict_hex_to_int,
    hex_to_number,
    get_size,
    convert_bytes,
    str2bool,
    flatten,
    flatten_list,
    flatten_dict,
    recursive_update_dict,
    dict_to_line,
    dict_none_to_zero,
    list_to_oneline_string,
    long_to_bytes,
    ordereddict_to_dict,
    extract_values_in_list,
    execute_function,
    influxdb_metrics_dict,
    metrics_key_push,
    dict2influxdb_line,
    rm_space,
    replace_ignore_char,
    replace_ignore_dict_kv,
    influx_key_value,
    split_every_n,
    class_extract_attr_list,
    append_zero,
    append_suffix,
    append_prefix,
    camel_case_to_space_case,
    camelcase_to_underscore,
    camel_case_to_lower_case,
    lower_case_to_camel_case,
    camel_case_to_upper_case,
    upper_case_to_camel_case,
    lower_case,
    upper_case,
    snake_case,
    shorten_text,
    truncate_float,
    truncate_decimal,
    remove_zero,
    remove_tags,
    remove_ascii_color_codes,
    json_to_hexadecimal,
    hexadecimal_to_json,
)

from .constants import (
    const
)

from .generator import (
    Null,
    Counter,
    GenMultiMetrics,
    id_generator,
    uuid_generator,
    generate_number_list,
    generate_json_rpc,
    json_rpc,
    parse_regex_number_list,
    token_hex,
    token_bytes,
    random_private_key,
    random_token_address,
    increase_token_address,
    increase_hex,
    increase_number,
    hexadecimal,
    decimal,
    uuid,

)

from .date_utils import (
    TimeCalculator,
    convert_unix_timestamp,
    get_range_day_of_month,
    todaydate,
    format_seconds_to_hhmmss,
    timestamp_to_string,
    second_to_dayhhmm,
)

